@@ -97,6 +97,8 @@ class BaseSession(ObservableMixin):
         }
 
         # session authentication information
+        self._realm = None
+        self._session_id = None
         self._authid = None
         self._authrole = None
         self._authmethod = None
@@ -108,6 +110,30 @@ class BaseSession(ObservableMixin):
         # generator for WAMP request IDs
         self._request_id_gen = IdGenerator()
 
+    @property
+    def realm(self):
+        return self._realm
+
+    @property
+    def session_id(self):
+        return self._session_id
+
+    @property
+    def authid(self):
+        return self._authid
+
+    @property
+    def authrole(self):
+        return self._authrole
+
+    @property
+    def authmethod(self):
+        return self._authmethod
+
+    @property
+    def authprovider(self):
+        return self._authprovider
+
     def define(self, exception, error=None):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.define`
@@ -289,7 +315,10 @@ class ApplicationSession(BaseSession):
         Implements :func:`autobahn.wamp.interfaces.ISession`
         """
         BaseSession.__init__(self)
-        self.config = config or types.ComponentConfig(realm=u"default")
+        self.config = config or types.ComponentConfig(realm=u"realm1")
+
+        # set client role features supported and announced
+        self._session_roles = role.DEFAULT_CLIENT_ROLES
 
         self._transport = None
         self._session_id = None
@@ -388,7 +417,7 @@ class ApplicationSession(BaseSession):
 
         # send HELLO message to router
         msg = message.Hello(realm=realm,
-                            roles=role.DEFAULT_CLIENT_ROLES,
+                            roles=self._session_roles,
                             authmethods=authmethods,
                             authid=authid,
                             authrole=authrole,
@@ -425,11 +454,14 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
         """
-        if False and isinstance(fail.value, exception.ApplicationError):
-            self.log.error(fail.value.error_message())
+        if isinstance(fail.value, exception.ApplicationError):
+            self.log.warn('{klass}.onUserError(): "{msg}"',
+                          klass=self.__class__.__name__,
+                          msg=fail.value.error_message())
         else:
             self.log.error(
-                u'{msg}: {traceback}',
+                '{klass}.onUserError(): "{msg}"\n{traceback}',
+                klass=self.__class__.__name__,
                 msg=msg,
                 traceback=txaio.failure_format_traceback(fail),
             )
@@ -484,17 +516,20 @@ class ApplicationSession(BaseSession):
 
                     if msg.realm:
                         self._realm = msg.realm
-
                     self._session_id = msg.session
+                    self._authid = msg.authid
+                    self._authrole = msg.authrole
+                    self._authmethod = msg.authmethod
+                    self._authprovider = msg.authprovider
                     self._router_roles = msg.roles
 
                     details = SessionDetails(
                         realm=self._realm,
                         session=self._session_id,
-                        authid=msg.authid,
-                        authrole=msg.authrole,
-                        authmethod=msg.authmethod,
-                        authprovider=msg.authprovider,
+                        authid=self._authid,
+                        authrole=self._authrole,
+                        authmethod=self._authmethod,
+                        authprovider=self._authprovider,
                         authextra=msg.authextra,
                         resumed=msg.resumed,
                         resumable=msg.resumable,
@@ -661,7 +696,7 @@ class ApplicationSession(BaseSession):
                         invoke_kwargs = msg.kwargs if msg.kwargs else dict()
 
                         if handler.details_arg:
-                            invoke_kwargs[handler.details_arg] = types.EventDetails(subscription, msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, retained=msg.retained, enc_algo=msg.enc_algo)
+                            invoke_kwargs[handler.details_arg] = types.EventDetails(subscription, msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, retained=msg.retained, enc_algo=msg.enc_algo, forward_for=msg.forward_for)
 
                         # FIXME: https://github.com/crossbario/autobahn-python/issues/764
                         def _success(_):
@@ -787,14 +822,11 @@ class ApplicationSession(BaseSession):
                             else:
                                 kw = msg.kwargs or dict()
                                 args = msg.args or tuple()
-                                try:
-                                    # XXX what if on_progress returns a Deferred/Future?
-                                    call_request.options.on_progress(*args, **kw)
-                                except Exception:
-                                    try:
-                                        self.onUserError(txaio.create_failure(), "While firing on_progress")
-                                    except:
-                                        pass
+
+                                def _error(fail):
+                                    self.onUserError(fail, "While firing on_progress")
+                                prog_d = txaio.as_future(call_request.options.on_progress, *args, **kw)
+                                txaio.add_callbacks(prog_d, None, _error)
 
                     else:
                         # process final call result
@@ -970,6 +1002,10 @@ class ApplicationSession(BaseSession):
                                         reply = message.Yield(msg.request,
                                                               args=[res])
 
+                                if self._transport is None:
+                                    self.log.debug('Skipping result of "{}", request {} because transport disconnected.'.format(registration.procedure, msg.request))
+                                    return
+
                                 try:
                                     self._transport.send(reply)
                                 except SerializationError as e:
@@ -1017,16 +1053,14 @@ class ApplicationSession(BaseSession):
             elif isinstance(msg, message.Interrupt):
 
                 if msg.request not in self._invocations:
-                    raise ProtocolError("INTERRUPT received for non-pending invocation {0}".format(msg.request))
+                    # raise ProtocolError("INTERRUPT received for non-pending invocation {0}".format(msg.request))
+                    self.log.debug('INTERRUPT received for non-pending invocation {request}', request=msg.request)
                 else:
                     invoked = self._invocations[msg.request]
-                    try:
-                        txaio.cancel(invoked.on_reply)
-                    finally:
-                        try:
-                            del self._invocations[msg.request]
-                        except KeyError:
-                            pass  # why? is the .cancel() handling cleaning this out somehow?
+                    # this will result in a CancelledError which will
+                    # be captured by the error handler around line 979
+                    # to delete the invocation..
+                    txaio.cancel(invoked.on_reply)
 
             elif isinstance(msg, message.Registered):
 
@@ -1244,11 +1278,12 @@ class ApplicationSession(BaseSession):
             msg = wamp.message.Goodbye(reason=reason, message=message)
             self._transport.send(msg)
             self._goodbye_sent = True
-            # deferred that fires when transport actually hits CLOSED
-            is_closed = self._transport is None or self._transport.is_closed
-            return is_closed
         else:
-            raise SessionNotReady(u"session was alread requested to leave")
+            self.log.warn('session was already requested to leave - not sending GOODBYE again')
+
+        is_closed = self._transport is None or self._transport.is_closed
+
+        return is_closed
 
     @public
     def onDisconnect(self):
@@ -1789,7 +1824,7 @@ class ApplicationSessionFactory(object):
         :param config: The default component configuration.
         :type config: instance of :class:`autobahn.wamp.types.ComponentConfig`
         """
-        self.config = config or types.ComponentConfig(realm=u"default")
+        self.config = config or types.ComponentConfig(realm=u"realm1")
 
     def __call__(self):
         """
