@@ -26,8 +26,10 @@
 
 from __future__ import absolute_import
 
+import os
 import six
 import struct
+import platform
 
 from autobahn.wamp.interfaces import IObjectSerializer, ISerializer
 from autobahn.wamp.exception import ProtocolError
@@ -40,6 +42,10 @@ __all__ = ['Serializer',
            'JsonSerializer']
 
 
+SERID_TO_OBJSER = {}
+SERID_TO_SER = {}
+
+
 class Serializer(object):
     """
     Base class for WAMP serializers. A WAMP serializer is the core glue between
@@ -79,7 +85,6 @@ class Serializer(object):
 
     def __init__(self, serializer):
         """
-        Constructor.
 
         :param serializer: The object serializer to use for WAMP wire-level serialization.
         :type serializer: An object that implements :class:`autobahn.interfaces.IObjectSerializer`.
@@ -98,100 +103,114 @@ class Serializer(object):
         """
         if isBinary is not None:
             if isBinary != self._serializer.BINARY:
-                raise ProtocolError("invalid serialization of WAMP message (binary {0}, but expected {1})".format(isBinary, self._serializer.BINARY))
-
+                raise ProtocolError(
+                    "invalid serialization of WAMP message (binary {0}, but expected {1})".format(isBinary,
+                                                                                                  self._serializer.BINARY))
         try:
             raw_msgs = self._serializer.unserialize(payload)
         except Exception as e:
-            raise ProtocolError("invalid serialization of WAMP message ({0})".format(e))
+            raise ProtocolError("invalid serialization of WAMP message: {0} {1}".format(type(e).__name__, e))
 
-        msgs = []
+        if self._serializer.NAME == u'flatbuffers':
+            msgs = raw_msgs
+        else:
+            msgs = []
+            for raw_msg in raw_msgs:
 
-        for raw_msg in raw_msgs:
+                if type(raw_msg) != list:
+                    raise ProtocolError("invalid type {0} for WAMP message".format(type(raw_msg)))
 
-            if type(raw_msg) != list:
-                raise ProtocolError("invalid type {0} for WAMP message".format(type(raw_msg)))
+                if len(raw_msg) == 0:
+                    raise ProtocolError(u"missing message type in WAMP message")
 
-            if len(raw_msg) == 0:
-                raise ProtocolError(u"missing message type in WAMP message")
+                message_type = raw_msg[0]
 
-            message_type = raw_msg[0]
+                if type(message_type) not in six.integer_types:
+                    # CBOR doesn't roundtrip number types
+                    # https://bitbucket.org/bodhisnarkva/cbor/issues/6/number-types-dont-roundtrip
+                    raise ProtocolError("invalid type {0} for WAMP message type".format(type(message_type)))
 
-            if type(message_type) not in six.integer_types:
-                # CBOR doesn't roundtrip number types
-                # https://bitbucket.org/bodhisnarkva/cbor/issues/6/number-types-dont-roundtrip
-                raise ProtocolError("invalid type {0} for WAMP message type".format(type(message_type)))
+                Klass = self.MESSAGE_TYPE_MAP.get(message_type)
 
-            Klass = self.MESSAGE_TYPE_MAP.get(message_type)
+                if Klass is None:
+                    raise ProtocolError("invalid WAMP message type {0}".format(message_type))
 
-            if Klass is None:
-                raise ProtocolError("invalid WAMP message type {0}".format(message_type))
+                # this might again raise `ProtocolError` ..
+                msg = Klass.parse(raw_msg)
 
-            # this might again raise `ProtocolError` ..
-            msg = Klass.parse(raw_msg)
-
-            msgs.append(msg)
+                msgs.append(msg)
 
         return msgs
 
 
 # JSON serialization is always supported
-
-import json
-import base64
+_USE_UJSON = 'AUTOBAHN_USE_UJSON' in os.environ
+if _USE_UJSON:
+    try:
+        import ujson
+        _USE_UJSON = True
+    except ImportError:
+        import json
+        _USE_UJSON = False
+else:
+    import json
 
 
-class _WAMPJsonEncoder(json.JSONEncoder):
+if _USE_UJSON:
+    # ujson doesn't support plugging into the JSON string parsing machinery ..
+    print('WARNING: Autobahn is using ujson accelerated JSON module - will run faster,\nbut only on CPython and will loose ability to transport binary payload transparently!')
+    _loads = ujson.loads
+    _dumps = ujson.dumps
+    _json = ujson
+else:
+    # print('Notice: Autobahn is using json built-in standard library module for JSON serialization')
+    import base64
 
-    def default(self, obj):
-        if isinstance(obj, six.binary_type):
-            return u'\x00' + base64.b64encode(obj).decode('ascii')
-        else:
-            return json.JSONEncoder.default(self, obj)
+    class _WAMPJsonEncoder(json.JSONEncoder):
 
+        def default(self, obj):
+            if isinstance(obj, six.binary_type):
+                return u'\x00' + base64.b64encode(obj).decode('ascii')
+            else:
+                return json.JSONEncoder.default(self, obj)
 
-#
-# the following is a hack. see http://bugs.python.org/issue29992
-#
+    #
+    # the following is a hack. see http://bugs.python.org/issue29992
+    #
 
-from json import scanner
-from json.decoder import scanstring
+    from json import scanner
+    from json.decoder import scanstring
 
+    def _parse_string(*args, **kwargs):
+        s, idx = scanstring(*args, **kwargs)
+        if s and s[0] == u'\x00':
+            s = base64.b64decode(s[1:])
+        return s, idx
 
-def _parse_string(*args, **kwargs):
-    s, idx = scanstring(*args, **kwargs)
-    if s and s[0] == u'\x00':
-        s = base64.b64decode(s[1:])
-    return s, idx
+    class _WAMPJsonDecoder(json.JSONDecoder):
 
+        def __init__(self, *args, **kwargs):
+            json.JSONDecoder.__init__(self, *args, **kwargs)
+            self.parse_string = _parse_string
 
-class _WAMPJsonDecoder(json.JSONDecoder):
+            # we need to recreate the internal scan function ..
+            self.scan_once = scanner.py_make_scanner(self)
 
-    def __init__(self, *args, **kwargs):
-        json.JSONDecoder.__init__(self, *args, **kwargs)
-        self.parse_string = _parse_string
+            # .. and we have to explicitly use the Py version,
+            # not the C version, as the latter won't work
+            # self.scan_once = scanner.make_scanner(self)
 
-        # we need to recreate the internal scan function ..
-        self.scan_once = scanner.py_make_scanner(self)
+    def _loads(s):
+        return json.loads(s, cls=_WAMPJsonDecoder)
 
-        # .. and we have to explicitly use the Py version,
-        # not the C version, as the latter won't work
-        # self.scan_once = scanner.make_scanner(self)
+    def _dumps(obj):
+        return json.dumps(obj,
+                          separators=(',', ':'),
+                          ensure_ascii=False,
+                          sort_keys=False,
+                          cls=_WAMPJsonEncoder)
 
-
-def _loads(s):
-    return json.loads(s, cls=_WAMPJsonDecoder)
-
-
-def _dumps(obj):
-    return json.dumps(obj,
-                      separators=(',', ':'),
-                      ensure_ascii=False,
-                      sort_keys=False,
-                      cls=_WAMPJsonEncoder)
-
-
-_json = json
+    _json = json
 
 
 class JsonObjectSerializer(object):
@@ -201,6 +220,8 @@ class JsonObjectSerializer(object):
     The JSON module used (now only stdlib).
     """
 
+    NAME = u'json'
+
     BINARY = False
 
     def __init__(self, batched=False):
@@ -238,6 +259,7 @@ class JsonObjectSerializer(object):
 
 
 IObjectSerializer.register(JsonObjectSerializer)
+SERID_TO_OBJSER[JsonObjectSerializer.NAME] = JsonObjectSerializer
 
 
 class JsonSerializer(Serializer):
@@ -273,20 +295,44 @@ class JsonSerializer(Serializer):
 
 
 ISerializer.register(JsonSerializer)
+SERID_TO_SER[JsonSerializer.SERIALIZER_ID] = JsonSerializer
 
 
-# MsgPack serialization depends on the `u-msgpack` package being available
-# https://pypi.python.org/pypi/u-msgpack-python
-# https://github.com/vsergeev/u-msgpack-python
-#
-try:
-    import umsgpack
-except ImportError:
-    pass
+_HAS_MSGPACK = False
+if platform.python_implementation() == u'CPython':
+    try:
+        # on CPython, use an impl. with native extension:
+        # https://pypi.org/project/msgpack/
+        # https://github.com/msgpack/msgpack-python
+        import msgpack
+    except ImportError:
+        pass
+    else:
+        _HAS_MSGPACK = True
+        _packb = lambda obj: msgpack.packb(obj, use_bin_type=True)  # noqa
+        _unpackb = lambda data: msgpack.unpackb(data, raw=False)  # noqa
+        # print('Notice: Autobahn is using msgpack library (with native extension, best on CPython) for MessagePack serialization')
 else:
+    try:
+        # on PyPy in particular, use a pure python impl.:
+        # https://pypi.python.org/pypi/u-msgpack-python
+        # https://github.com/vsergeev/u-msgpack-python
+        import umsgpack
+    except ImportError:
+        pass
+    else:
+        _HAS_MSGPACK = True
+        _packb = umsgpack.packb
+        _unpackb = umsgpack.unpackb
+        # print('Notice: Autobahn is using umsgpack library (pure Python, best on PyPy) for MessagePack serialization')
+
+
+if _HAS_MSGPACK:
 
     class MsgPackObjectSerializer(object):
 
+        NAME = u'msgpack'
+
         BINARY = True
         """
         Flag that indicates whether this serializer needs a binary clean transport.
@@ -304,7 +350,7 @@ else:
             """
             Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
             """
-            data = umsgpack.packb(obj)
+            data = _packb(obj)
             if self._batched:
                 return struct.pack("!L", len(data)) + data
             else:
@@ -331,7 +377,7 @@ else:
                     data = payload[i + 4:i + 4 + l]
 
                     # append parsed raw message
-                    msgs.append(umsgpack.unpackb(data))
+                    msgs.append(_unpackb(data))
 
                     # advance until everything consumed
                     i = i + 4 + l
@@ -341,12 +387,13 @@ else:
                 return msgs
 
             else:
-                unpacked = umsgpack.unpackb(payload)
+                unpacked = _unpackb(payload)
                 return [unpacked]
 
     IObjectSerializer.register(MsgPackObjectSerializer)
 
     __all__.append('MsgPackObjectSerializer')
+    SERID_TO_OBJSER[MsgPackObjectSerializer.NAME] = MsgPackObjectSerializer
 
     class MsgPackSerializer(Serializer):
 
@@ -380,22 +427,44 @@ else:
                 self.SERIALIZER_ID = u"msgpack.batched"
 
     ISerializer.register(MsgPackSerializer)
+    SERID_TO_SER[MsgPackSerializer.SERIALIZER_ID] = MsgPackSerializer
 
     __all__.append('MsgPackSerializer')
 
 
-# CBOR serialization depends on the `cbor` package being available
-# https://pypi.python.org/pypi/cbor
-# https://bitbucket.org/bodhisnarkva/cbor
-#
-try:
-    import cbor
-except ImportError:
-    pass
+_HAS_CBOR = False
+if 'AUTOBAHN_USE_CBOR2' in os.environ:
+    try:
+        # https://pypi.org/project/cbor2/
+        # https://github.com/agronholm/cbor2
+        import cbor2
+    except ImportError:
+        pass
+    else:
+        _HAS_CBOR = True
+        _cbor_loads = cbor2.loads
+        _cbor_dumps = cbor2.dumps
+        # print('Notice: Autobahn is using cbor2 library for CBOR serialization')
 else:
+    try:
+        # https://pypi.python.org/pypi/cbor
+        # https://bitbucket.org/bodhisnarkva/cbor
+        import cbor
+    except ImportError:
+        pass
+    else:
+        _HAS_CBOR = True
+        _cbor_loads = cbor.loads
+        _cbor_dumps = cbor.dumps
+        # print('Notice: Autobahn is using cbor library for CBOR serialization')
+
+
+if _HAS_CBOR:
 
     class CBORObjectSerializer(object):
 
+        NAME = u'cbor'
+
         BINARY = True
         """
         Flag that indicates whether this serializer needs a binary clean transport.
@@ -414,7 +483,7 @@ else:
             """
             Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
             """
-            data = cbor.dumps(obj)
+            data = _cbor_dumps(obj)
             if self._batched:
                 return struct.pack("!L", len(data)) + data
             else:
@@ -441,7 +510,7 @@ else:
                     data = payload[i + 4:i + 4 + l]
 
                     # append parsed raw message
-                    msgs.append(cbor.loads(data))
+                    msgs.append(_cbor_loads(data))
 
                     # advance until everything consumed
                     i = i + 4 + l
@@ -451,10 +520,11 @@ else:
                 return msgs
 
             else:
-                unpacked = cbor.loads(payload)
+                unpacked = _cbor_loads(payload)
                 return [unpacked]
 
     IObjectSerializer.register(CBORObjectSerializer)
+    SERID_TO_OBJSER[CBORObjectSerializer.NAME] = CBORObjectSerializer
 
     __all__.append('CBORObjectSerializer')
 
@@ -490,6 +560,7 @@ else:
                 self.SERIALIZER_ID = u"cbor.batched"
 
     ISerializer.register(CBORSerializer)
+    SERID_TO_SER[CBORSerializer.SERIALIZER_ID] = CBORSerializer
 
     __all__.append('CBORSerializer')
 
@@ -502,9 +573,12 @@ try:
 except ImportError:
     pass
 else:
+    # print('Notice: Autobahn is using ubjson module for UBJSON serialization')
 
     class UBJSONObjectSerializer(object):
 
+        NAME = u'ubjson'
+
         BINARY = True
         """
         Flag that indicates whether this serializer needs a binary clean transport.
@@ -563,6 +637,7 @@ else:
                 return [unpacked]
 
     IObjectSerializer.register(UBJSONObjectSerializer)
+    SERID_TO_OBJSER[UBJSONObjectSerializer.NAME] = UBJSONObjectSerializer
 
     __all__.append('UBJSONObjectSerializer')
 
@@ -598,5 +673,129 @@ else:
                 self.SERIALIZER_ID = u"ubjson.batched"
 
     ISerializer.register(UBJSONSerializer)
+    SERID_TO_SER[UBJSONSerializer.SERIALIZER_ID] = UBJSONSerializer
 
     __all__.append('UBJSONSerializer')
+
+
+_HAS_FLATBUFFERS = False
+try:
+    import flatbuffers  # noqa
+    from autobahn.wamp import message_fbs
+except ImportError:
+    pass
+else:
+    _HAS_FLATBUFFERS = True
+
+
+if _HAS_FLATBUFFERS:
+
+    class FlatBuffersObjectSerializer(object):
+
+        NAME = u'flatbuffers'
+
+        BINARY = True
+        """
+        Flag that indicates whether this serializer needs a binary clean transport.
+        """
+
+        MESSAGE_TYPE_MAP = {
+            message_fbs.MessageType.EVENT: (message_fbs.Event, message.Event),
+            message_fbs.MessageType.PUBLISH: (message_fbs.Publish, message.Publish),
+        }
+
+        def __init__(self, batched=False):
+            """
+
+            :param batched: Flag that controls whether serializer operates in batched mode.
+            :type batched: bool
+            """
+            assert not batched, 'WAMP-FlatBuffers serialization does not support message batching currently'
+            self._batched = batched
+
+        def serialize(self, obj):
+            """
+            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
+            """
+            raise NotImplementedError()
+
+        def unserialize(self, payload):
+            """
+            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.unserialize`
+            """
+            union_msg = message_fbs.Message.Message.GetRootAsMessage(payload, 0)
+            msg_type = union_msg.MsgType()
+
+            if msg_type in self.MESSAGE_TYPE_MAP:
+                fbs_klass, wamp_klass = self.MESSAGE_TYPE_MAP[msg_type]
+                fbs_msg = fbs_klass()
+                _tab = union_msg.Msg()
+                fbs_msg.Init(_tab.Bytes, _tab.Pos)
+                msg = wamp_klass(from_fbs=fbs_msg)
+                return [msg]
+            else:
+                raise NotImplementedError('message type {} not yet implemented for WAMP-FlatBuffers'.format(msg_type))
+
+    IObjectSerializer.register(FlatBuffersObjectSerializer)
+
+    __all__.append('FlatBuffersObjectSerializer')
+    SERID_TO_OBJSER[FlatBuffersObjectSerializer.NAME] = FlatBuffersObjectSerializer
+
+    class FlatBuffersSerializer(Serializer):
+
+        SERIALIZER_ID = u"flatbuffers"
+        """
+        ID used as part of the WebSocket subprotocol name to identify the
+        serializer with WAMP-over-WebSocket.
+        """
+
+        RAWSOCKET_SERIALIZER_ID = 5
+        """
+        ID used in lower four bits of second octet in RawSocket opening
+        handshake identify the serializer with WAMP-over-RawSocket.
+        """
+
+        MIME_TYPE = u"application/x-flatbuffers"
+        """
+        MIME type announced in HTTP request/response headers when running
+        WAMP-over-Longpoll HTTP fallback.
+        """
+
+        def __init__(self, batched=False):
+            """
+
+            :param batched: Flag to control whether to put this serialized into batched mode.
+            :type batched: bool
+            """
+            Serializer.__init__(self, FlatBuffersObjectSerializer(batched=batched))
+            if batched:
+                self.SERIALIZER_ID = u"flatbuffers.batched"
+
+    ISerializer.register(FlatBuffersSerializer)
+    SERID_TO_SER[FlatBuffersSerializer.SERIALIZER_ID] = FlatBuffersSerializer
+
+    __all__.append('FlatBuffersSerializer')
+
+
+def create_transport_serializer(serializer_id):
+    batched = False
+    if '.' in serializer_id:
+        l = serializer_id.split('.')
+        serializer_id = l[0]
+        if len(l) > 1 and l[1] == 'batched':
+            batched = True
+
+    if serializer_id in SERID_TO_SER:
+        return SERID_TO_SER[serializer_id](batched=batched)
+    else:
+        raise RuntimeError('could not create serializer for "{}" (available: {})'.format(serializer_id, sorted(SERID_TO_SER.keys())))
+
+
+def create_transport_serializers(transport):
+    """
+    Create a list of serializers to use with a WAMP protocol factory.
+    """
+    serializers = []
+    for serializer_id in transport.serializers:
+        serializers.append(create_transport_serializer(serializer_id))
+    return serializers
